/* GNU POP3 - a small, fast, and efficient POP3 daemon
   Copyright (C) 1999 Jakob 'sparky' Kaivo <jkaivo@nodomainname.net>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  */

#include "gnu-pop3d.h"

/* Check if a username exists in APOP password file
   returns pointer to password if found, otherwise NULL */
char *
pop3_apopuser (const char *user)
{
  char *password, *usr;
  char buf[POP_MAXCMDLEN];
  struct stat st;

#ifdef BDB2
  int errno;
  DB *dbp;
  DBT key, data;

  if (stat (APOP_PASSFILE ".db", &st) != -1)
#else
  FILE *apop_file;

  if (stat (APOP_PASSFILE ".passwd", &st) != -1)
#endif
	if ((st.st_mode & 0777) != 0600)
	  {
		syslog (LOG_INFO, "Bad permissions on APOP password file");
		return NULL;
	  }
#ifdef BDB2
  if ((errno = db_open (APOP_PASSFILE ".db", DB_HASH, DB_RDONLY, 0600, NULL, NULL, &dbp)) != 0)
	{
	  syslog (LOG_ERR, "Unable to open APOP database: %s", strerror (errno));
	  return NULL;
	}
  memset (&key, 0, sizeof (DBT));
  memset (&data, 0, sizeof (DBT));

  strncpy (buf, user, sizeof (buf));
  key.data = buf;
  key.size = strlen (user);
  if ((errno = dbp->get (dbp, NULL, &key, &data, 0)) != 0)
	{
	  syslog (LOG_ERR, "db_get error: %s", strerror (errno));
	  dbp->close (dbp, 0);
	  return NULL;
	}
  if ((password = malloc (sizeof (char) * data.size)) == NULL)
	{
	  dbp->close (dbp, 0);
	  return NULL;
	}
  sprintf (password, "%.*s", (int) data.size, (char *) data.data);
  dbp->close (dbp, 0);
  return password;
#else /* !BDBD2 */
  if ((apop_file = fopen (APOP_PASSFILE ".passwd", "r")) == NULL)
	{
	  syslog (LOG_INFO, "Unable to open APOP password file");
	  return NULL;
	}
  if ((password = malloc (sizeof (char) * APOP_DIGEST)) == NULL)
	{
	  fclose (apop_file);
	  pop3_abquit (ERR_NO_MEM);
	}
  password[0] = '\0';
  if ((usr = malloc (sizeof (char) * 128)) == NULL)
	{
	  fclose (apop_file);
	  pop3_abquit (ERR_NO_MEM);
	}
  usr[0] = '\0';

  while (fgets (buf, sizeof (buf), apop_file) != NULL)
	{

	  /* skip comments, blank lines */
	  if (*buf == '#' || *buf == '\n' || *buf == '\0')
		{
		  continue;
		}

	  if (sscanf (buf, "%[^:]:%127s", usr, password) != 2)
		{
		  syslog (LOG_ERR, "%s: Line format in %s.passwd file incorrect\n", buf, APOP_PASSFILE);
		  continue;
		}

	  if (strcmp (usr, user) == 0)
		break;
	}

  fclose (apop_file);
  free (usr);
  if (strlen (password) == 0)
	{
	  free (password);
	  password = NULL;
	}
  return password;
#endif /* BDB2 */
}

int
pop3_apop (const char *arg)
{
  char *tmp, *user_digest, *password;
  struct passwd *pw = NULL;
  char buf[POP_MAXCMDLEN];
  struct md5_ctx md5context;
  unsigned char md5digest[16];
  int i;

  if (state != AUTHORIZATION)
	return ERR_WRONG_STATE;
  if (strlen (arg) == 0)
	return ERR_BAD_ARGS;

  tmp = pop3_cmd (arg);
  if (strlen (tmp) > (POP_MAXCMDLEN - APOP_DIGEST))
	{
	  free (tmp);
	  return ERR_BAD_ARGS;
	}
  if ((username = strdup (tmp)) == NULL)
	pop3_abquit (ERR_NO_MEM);
  free (tmp);

#ifdef USE_VIRTUAL
  virtualdomain = pop3_splitname (username, useroptions);
#ifdef DEFAULT_DOMAIN
  if (virtualdomain)
	if (strcasecmp (DEFAULT_DOMAIN, virtualdomain) == 0)
	  {							/* fallback to normal */
		free (virtualdomain);
		virtualdomain = NULL;
	  }
#endif /* DEFAULT_DOMAIN */
  /* syslog(LOG_INFO, "APOP user %s AT %s trying to log in", username, (virtualdomain != NULL)?virtualdomain:"()"); */
#endif /* USE_VIRTUAL */

  tmp = pop3_args (arg);
  if (strlen (tmp) > APOP_DIGEST)
	{
	  free (tmp);
	  return ERR_BAD_ARGS;
	}
  if ((user_digest = strdup (tmp)) == NULL)
	pop3_abquit (ERR_NO_MEM);
  free (tmp);

#ifdef USE_VIRTUAL
  if (virtualdomain)
	{
	  if ((pw = getvirtualpwnam (username, virtualdomain)) == NULL)
		return ERR_BAD_LOGIN;
	  if (!pw->pw_passwd || !strcmp (pw->pw_shell, "encrypted"))
		{
		  pop3_freepw (pw);
		  return ERR_BAD_LOGIN;
		}
	  password = pw->pw_passwd;	/* must be unencrypted for APOP */
	}
  else
	{
#endif /* USE_VIRTUAL */

	  if ((password = pop3_apopuser (username)) == NULL)
		return ERR_BAD_LOGIN;

#ifdef USE_VIRTUAL
	}
#endif /* USE_VIRTUAL */

  md5_init_ctx (&md5context);
  md5_process_bytes (md5shared, strlen (md5shared), &md5context);
  md5_process_bytes (password, strlen (password), &md5context);
  md5_finish_ctx (&md5context, md5digest);

#ifdef USE_VIRTUAL
  if (!virtualdomain)
	free (password);
#else
  free (password);
#endif

  tmp = buf;
  for (i = 0; i < 16; i++, tmp += 2)
	sprintf (tmp, "%02x", md5digest[i]);

  *tmp++ = '\0';

  if (strcmp (user_digest, buf))
	return ERR_BAD_LOGIN;

#ifdef USE_VIRTUAL
  if (virtualdomain == NULL)
#endif
	if ((pw = getpwnam (username)) == NULL)
	  return ERR_BAD_LOGIN;

  return pop3_finishlogin (pw);
}
